"""
This module contains a simple simpy.io based webserver with websockets.

"""
import json
import logging
import mimetypes
import os.path

from simpy.io import select as backend
from simpy.io.http import Service
from simpy.io.websocket import WebSocket


logger = logging.getLogger(__name__)

UPDATE_INTERVAL = 0.25


class Server(object):
    def __init__(self, env, server_sock):
        self.env = env
        self.server_sock = server_sock

        self.basedir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                    'html'))
        """Base directory for static files."""

        self.topology_ready = env.event()
        self.topology = None
        self.data_buf = None
        self.wait_for_update = []  # List of events to be triggered on updates
        self._reset_data_buf()

        self.env.process(self._broadcast_update())
        self.env.process(self._serve())

    def _broadcast_update(self):
        while True:
            yield self.env.timeout(UPDATE_INTERVAL)
            new_data = self._reset_data_buf()
            msg = json.dumps(['update_data', new_data])
            for evt in self.wait_for_update:
                evt.succeed(msg)
            self.wait_for_update = []

    def _serve(self):
        """Webserver main process."""
        while True:
            sock = yield self.server_sock.accept()
            service = Service(sock)
            self.env.process(self.handler(service))

    def handler(self, service):
        """Handle client requests.

        Either serve static files or create a websocket.

        """
        try:
            while True:
                request = yield service.recv()
                yield request.read(all=True)

                uri = request.uri
                if uri.endswith('/'):
                    uri += 'index.html'

                if uri == '/websocket':
                    self.env.process(self.websock(service, request))
                    break

                try:
                    ctype, data = self.serve_static(uri)
                except ValueError:
                    yield request.respond(404, {
                        'content-type': 'text/plain; charset=utf-8',
                    }, data=b'Not found')
                else:
                    yield request.respond(200, {
                        'content-type': ctype,
                    }, data=data)

        except ConnectionError:
            logger.warn('socket ConnectionError in "Server.handler()"')

    def websock(self, service, request):
        """Process for websocket connections."""
        excess_data = service.decommission()
        socket = WebSocket(service.sock)
        socket.configure(False, headers=request.headers)

        try:
            msg = yield socket.read()
            assert msg == 'get_topology'
            yield self.topology_ready
            yield socket.write(json.dumps(['setup_topology', self.topology]))

            while True:
                evt_new_data = self.env.event()
                self.wait_for_update.append(evt_new_data)
                msg = yield evt_new_data
                yield socket.write(msg)

        except ConnectionError:
            logger.warn('websocket ConnectionError in "Server.websock()"')

    def serve_static(self, uri):
        """Try to read and return a static file and its mime type."""
        req_path = os.path.abspath(os.path.join(self.basedir, uri.lstrip('/')))
        if not req_path.startswith(self.basedir):
            raise ValueError
        if not os.path.isfile(req_path):
            raise ValueError

        content_type = mimetypes.guess_type(req_path)[0]
        if content_type.startswith('text/'):
            content_type = '%s; charset=utf-8' % content_type
        return content_type, open(req_path, 'rb').read()

    def set_new_data(self, time, progress, node_data):
        self.data_buf['time'] = time
        self.data_buf['progress'] = progress
        self.data_buf['node_data'].append(node_data)

    def _reset_data_buf(self):
        data = self.data_buf
        self.data_buf = {
            'time': None,
            'progress': None,
            'node_data': [],
        }
        return data


if __name__ == '__main__':
    addr = ('localhost', 8000)

    env = backend.Environment()
    server_sock = backend.TCPSocket.server(env, addr)
    server = Server(env, server_sock)

    env.run()



from dateutil import tz
import json
import logging
import time

import arrow
import mosaik_api
import networkx as nx

from mosaik_web import server


logger = logging.getLogger('mosaik_web.mosaik')

meta = {
    'models': {
        'Topology': {
            'public': True,
            'params': [],
            'attrs': [],
        }
    },
}

# TODO: Document config file format
default_config = {
    'ignore_types': ['Topology'],
    'merge_types': ['Branch', 'Transformer'],
    'disable_heatmap': False,
    'timeline_hours': 24,
    'etypes': {},
}

DATE_FORMAT = 'YYYY-MM-DD HH:mm:ss'


class MosaikWeb(mosaik_api.Simulator):
    def __init__(self):
        super().__init__(meta)
        self.step_size = None
        self.server = None
        self.sid = None
        self.eid = None
        self.config = default_config

        self.start_rt = None  # Simulation start in wallclock time
        self.realtime_factor = None

    def configure(self, args, backend, env):
        """Start a wevserver for the visualization."""
        server_addr = mosaik_api._parse_addr(args['--serve'])
        server_sock = backend.TCPSocket.server(env, server_addr)
        self.server = server.Server(env, server_sock)

    def init(self, sid, start_date, step_size, config_file,
             realtime_factor=None):
        self.sid = sid
        dt = arrow.parser.DateTimeParser().parse(start_date, DATE_FORMAT)
        self.start_date = arrow.get(dt, tz.tzlocal()).isoformat()
        self.step_size = step_size
        self.realtime_factor = realtime_factor
        config = json.load(open(config_file))
        self.config.update(config)

        return self.meta

    def create(self, num, model):
        if num != 1 or self.eid is not None:
            raise ValueError('Can only one grid instance.')
        if model != 'Topology':
            raise ValueError('Unknown model: "%s"' % model)

        self.eid = 'topo'

        return [{'eid': self.eid, 'type': model, 'rel': []}]

    def step(self, time, inputs):
        if not self.server.topology:
            yield from self._build_topology()

        progress = yield self.mosaik.get_progress()

        etype_conf = self.config['etypes']
        req_attrs = {}
        for eid, etype in self.related_entities:
            if etype not in etype_conf:
                continue
            req_attrs[eid] = [etype_conf[etype]['attr']]
        data = yield self.mosaik.get_data(req_attrs)

        node_data = {}
        for node in self.server.topology['nodes']:
            node_id = node['name']
            try:
                attr = etype_conf[node['type']]['attr']
            except KeyError:
                val = 0
            else:
                val = data[node_id][attr]
            node_data[node_id] = {
                'value': val,
            }
        self.server.set_new_data(time, progress, node_data)

        if self.realtime_factor:
            self._sleep(time)

        return time + self.step_size

    def get_data(self, outputs):
        return {}

    def _build_topology(self):
        """Get all related entities, create the topology and set it to the
        web server."""
        logger.info('Creating topology ...')

        data = yield self.mosaik.get_related_entities()
        nxg = nx.Graph()
        nxg.add_nodes_from(data['nodes'].items())
        nxg.add_edges_from(data['edges'])

        # Required for get_data() calls.
        full_id = '%s.%s' % (self.sid, self.eid)
        self.related_entities = [(e, nxg.node[e]['type'])
                                 for e in nxg.neighbors(full_id)]

        self._clean_nx_graph(nxg)
        self.server.topology = self._make_d3js_topology(nxg)
        self.server.topology_ready.succeed()

        logger.info('Topology created')

    def _clean_nx_graph(self, nxg):
        """Remove and merge nodes and edges according to ``self.ignore_types``
        and ``self.merge_types``."""
        nxg.remove_nodes_from([n for n, d in nxg.node.items()
                               if d['type'] in self.config['ignore_types']])
        for node in [n for n, d in nxg.node.items()
                     if d['type'] in self.config['merge_types']]:
            new_edge = nxg.neighbors(node)
            assert len(new_edge) == 2, new_edge
            nxg.remove_node(node)
            nxg.add_edge(*new_edge)

    def _make_d3js_topology(self, nxg):
        """Create the topology for D3JS."""
        # We have to use two loops to make sure "node_idx" is filled for the
        # second one.
        topology = {
            'start_date': self.start_date,
            'update_interval': self.step_size,
            'timeline_hours': self.config['timeline_hours'],
            'disable_heatmap': self.config['disable_heatmap'],
            'etypes': self.config['etypes'],
            'nodes': [],
            'links': [],
        }
        node_idx = {}

        for node, attrs in nxg.node.items():
            node_idx[node] = len(topology['nodes'])
            type = attrs['type']
            topology['nodes'].append({
                'name': node,
                'type': type,
                'value': 0,
            })

        for source, target in nxg.edges():
            topology['links'].append({
                'source': node_idx[source],
                'target': node_idx[target],
                'length': 0,  # TODO: Add eddge data['length'],
            })

        return topology

    def _sleep(self, t_step):
        """Sleep to stay in sync with the real-time."""
        if not self.start_rt:
            # Set simulation start time
            self.start_rt = time.time()

        sim_time_passed = t_step + self.step_size
        rt_passed = sim_time_passed * self.realtime_factor
        sleep_until = self.start_rt + rt_passed
        sleep_t = sleep_until - time.time()
        if sleep_t < 0:
            logger.warn('Behind real time (%s seconds)!' % sleep_t)
        else:
            logger.debug('Gonna sleep for %s seconds.' % sleep_t)
            time.sleep(sleep_t)


def main():
    desc = 'Simple visualization for mosaik simulations'
    extra_opts = [
        '-s HOST:PORT, --serve=HOST:PORT    ',
        ('            Host and port for the webserver '
         '[default: 127.0.0.1:8000]'),
    ]
    mosaik_api.start_simulation(MosaikWeb(), desc, extra_opts)
